@@ -31,6 +31,7 @@
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <sys/poll.h>
+#include <sys/wait.h>
 #include <fcntl.h>
 #include <time.h>
 #include <stdarg.h>
@@ -159,7 +160,7 @@ static int timed_connect(int sock, const struct sockaddr *addr, unsigned int len
 	pfd[0].fd=sock;
 	pfd[0].events=POLLOUT;	
 	fcntl(sock, F_SETFL, O_NONBLOCK);
-  	ret=__libc_connect(sock, addr,  len);
+  	ret=true_connect(sock, addr,  len);
 //	printf("\nconnect ret=%d\n",ret);fflush(stdout);
   	if(ret==-1 && errno==EINPROGRESS)
    	{
@@ -295,7 +296,7 @@ static int tunnel_to(int sock, unsigned int ip, unsigned short port, proxy_type
 			
       			if (buff[0]!=5||(buff[1]!=0&&buff[1]!=2))
          			{
-         				if((buff[0]==5)&&(buff[1]==0xFF))
+        				if((buff[0]==0x05)&&(buff[1]==(char)0xFF))
              						return BLOCKED;
 						else
 							return SOCKET_ERROR;
@@ -374,33 +375,39 @@ static int tunnel_to(int sock, unsigned int ip, unsigned short port, proxy_type
 return SOCKET_ERROR;
 }
 
-static int start_chain(int *fd, proxy_data *pd)
+static int start_chain(int *fd, proxy_data *pd, char* begin_mark)
 {
 	struct sockaddr_in addr;
-	pd->ps=PLAY_STATE;
+	
 	*fd=socket(PF_INET,SOCK_STREAM,0);
 	if(*fd==-1)
 		goto error;
+	
+	proxychains_write_log("%s-<>-%s:%d-",
+				begin_mark,
+				inet_ntoa(*(struct in_addr*)&pd->ip),
+				htons(pd->port));
+	pd->ps=PLAY_STATE;
 	bzero(&addr,sizeof(addr));
 	addr.sin_family = AF_INET;
 	addr.sin_addr.s_addr = pd->ip;
 	addr.sin_port = pd->port;
 	if (timed_connect (*fd ,(struct sockaddr*)&addr,sizeof(addr))) {
 		pd->ps=DOWN_STATE;
-		goto error;
+		goto error1;
 	}
 	pd->ps=BUSY_STATE;
-	proxychains_write_log("..%s:%d..",
-				inet_ntoa(*(struct in_addr*)&pd->ip),
-				htons(pd->port));
 	return SUCCESS;
+error1:
+	proxychains_write_log("<--timeout\n");
 error:
 	if(*fd!=-1)
 		close(*fd);
 	return SOCKET_ERROR;
 }
 
-static proxy_data * select_proxy(select_type how, proxy_data *pd, int proxy_count, int *offset)
+static proxy_data * select_proxy(select_type how, 
+		proxy_data *pd, int proxy_count, int *offset)
 {
 	int i=0,k=0;
 	if(*offset>=proxy_count)
@@ -410,7 +417,8 @@ static proxy_data * select_proxy(select_type how, proxy_data *pd, int proxy_coun
 			srand(time(NULL));
 			do {
 				k++;
-				i = 0 + (int) (proxy_count*1.0*rand()/(RAND_MAX+1.0));
+				i = 0 + (int) (proxy_count*1.0*rand()/
+						(RAND_MAX+1.0));
 			} while (pd[i].ps!=PLAY_STATE && k<proxy_count*100 );
 		break;
 		case FIFOLY:
@@ -428,6 +436,15 @@ static proxy_data * select_proxy(select_type how, proxy_data *pd, int proxy_coun
 	return pd[i].ps==PLAY_STATE?&pd[i]:NULL;
 }
 
+
+static void release_all(proxy_data *pd, int proxy_count)
+{
+	int i;
+	for(i=0;i<proxy_count;i++)
+		pd[i].ps=PLAY_STATE;
+	return;
+}
+
 static void release_busy(proxy_data *pd, int proxy_count)
 {
 	int i;
@@ -452,40 +469,34 @@ static int calc_alive(proxy_data *pd, int proxy_count)
 static int chain_step(int ns, proxy_data *pfrom, proxy_data *pto)
 {
 	int retcode=-1;
-	retcode=tunnel_to(ns,pto->ip,pto->port,pfrom->pt,pfrom->user,pfrom->pass);
+	
+	proxychains_write_log("<>-%s:%d-", 
+			inet_ntoa(*(struct in_addr*)&pto->ip),
+			htons(pto->port));
+	retcode = 
+		tunnel_to(ns, pto->ip, pto->port, pfrom->pt, pfrom->user, 
+				pfrom->pass);
 	switch(retcode) {
 		case SUCCESS:
 			pto->ps=BUSY_STATE;
-			proxychains_write_log("..%s:%d..",
-				inet_ntoa(*(struct in_addr*)&pto->ip),
-				htons(pto->port));
 			break;
 		case BLOCKED:
 			pto->ps=BLOCKED_STATE;
-			proxychains_write_log("..access denied to..%s:%d\n",
-				inet_ntoa(*(struct in_addr*)&pto->ip),
-				htons(pto->port));
+			proxychains_write_log("<--denied\n");
 			close(ns);
 			break;
 		case SOCKET_ERROR:
 			pto->ps=DOWN_STATE;
-			proxychains_write_log("..can't connect to..%s:%d\n",
-				inet_ntoa(*(struct in_addr*)&pto->ip),
-				htons(pto->port));
+			proxychains_write_log("<--timeout\n");
 			close(ns);
 			break;
 	}
 	return retcode;
 }
 
-int connect_proxy_chain(
-	int sock,
-	unsigned int target_ip,
-	unsigned short target_port,
-	proxy_data *pd,
-	unsigned int proxy_count,
-	chain_type ct,
-	int max_chain )
+int connect_proxy_chain( int sock, unsigned int target_ip, 
+		unsigned short target_port, proxy_data *pd, 
+		unsigned int proxy_count, chain_type ct, int max_chain )
 {
 	proxy_data p4;
 	proxy_data *p1,*p2,*p3;
@@ -493,97 +504,230 @@ int connect_proxy_chain(
 	int offset=0;
 	int alive_count=0;
 	int curr_len=0;
+
+#define TP "<>"
+#define DT "|D-chain|"
+#define ST "|S-chain|"
+#define RT "|R-chain|"
+	
 	p3=&p4;
+
+again:
 	switch(ct)  {
 		case DYNAMIC_TYPE:
-			{
-			again_dynamic:
-				alive_count=calc_alive(pd,proxy_count);
-				offset=0;
-				proxychains_write_log("dynamic chain:..");
-				do {
-					if(!(p1=select_proxy(FIFOLY,pd,proxy_count,&offset)))
-						goto error_more;
-				} while(SUCCESS!=start_chain(&ns,p1) && offset<proxy_count);
-				for(;;) {
-					p2=select_proxy(FIFOLY,pd,proxy_count,&offset);
-					if(!p2)
-						break;
-					if(SUCCESS!=chain_step(ns,p1,p2))
-						goto again_dynamic;
-					p1=p2;
-				}
-				p3->ip=target_ip;
-				p3->port=target_port;
-				if(SUCCESS!=chain_step(ns,p1,p3))
-					goto again_dynamic;
-				proxychains_write_log("OK\n");
-			}
-			break;
-		case STRICT_TYPE:
-			{
-				alive_count=calc_alive(pd,proxy_count);
-				offset=0;
-				proxychains_write_log("strict chain:..");
-				if(!(p1=select_proxy(FIFOLY,pd,proxy_count,&offset)))
-					goto error_strict;
-				if(SUCCESS!=start_chain(&ns,p1))
-					goto error_strict;
-				while(offset<proxy_count) {
-					if(!(p2=select_proxy(FIFOLY,pd,proxy_count,&offset)))
-						goto error_strict;
-					if(SUCCESS!=chain_step(ns,p1,p2))
-						goto error;
-					p1=p2;
-				}
-				p3->ip=target_ip;
-				p3->port=target_port;
-				if(SUCCESS!=chain_step(ns,p1,p3))
-					goto error;
-				proxychains_write_log("OK\n");
-			}
-			break;
-		case RANDOM_TYPE:
-			again_random:
-			{
-				alive_count=calc_alive(pd,proxy_count);
-				if(alive_count<max_chain)
-					goto error_more;
-				curr_len=offset=0;
-				proxychains_write_log("random chain (%d):..",max_chain);
-				do {
-					if(!(p1=select_proxy(RANDOMLY,pd,proxy_count,&offset)))
-						goto error_more;
-				} while(SUCCESS!=start_chain(&ns,p1) && offset<max_chain);
-				while(++curr_len<max_chain) {
-					if(!(p2=select_proxy(RANDOMLY,pd,proxy_count,&offset)))
-						goto error_more;
-					if(SUCCESS!=chain_step(ns,p1,p2))
-						goto again_random;
-					p1=p2;
-				}
-				p3->ip=target_ip;
-				p3->port=target_port;
-				if(SUCCESS!=chain_step(ns,p1,p3))
-					goto again_random;
-				proxychains_write_log("OK\n");
-			}
+		alive_count=calc_alive(pd,proxy_count);
+		offset=0;
+		do {
+			if(!(p1=select_proxy(FIFOLY,pd,proxy_count,&offset)))
+				goto error_more;
+		} while(SUCCESS!=start_chain(&ns,p1,DT) && offset<proxy_count);
+		for(;;) {
+			p2=select_proxy(FIFOLY,pd,proxy_count,&offset);
+			if(!p2)
+				break;
+			if(SUCCESS!=chain_step(ns,p1,p2))
+				goto again;
+			p1=p2;
+		}
+		proxychains_write_log(TP);
+		p3->ip=target_ip;
+		p3->port=target_port;
+		if(SUCCESS!=chain_step(ns,p1,p3))
+			goto error;
+		break;
+
+	case STRICT_TYPE:
+		alive_count=calc_alive(pd,proxy_count);
+		offset=0;
+		if(!(p1=select_proxy(FIFOLY,pd,proxy_count,&offset)))
+			goto error_strict;
+		if(SUCCESS!=start_chain(&ns,p1,ST))
+			goto error_strict;
+		while(offset<proxy_count) {
+			if(!(p2=select_proxy(FIFOLY,pd,proxy_count,&offset)))
+				break;
+			if(SUCCESS!=chain_step(ns,p1,p2))
+				goto error_strict;
+			p1=p2;
+		}
+		proxychains_write_log(TP);
+		p3->ip=target_ip;
+		p3->port=target_port;
+		if(SUCCESS!=chain_step(ns,p1,p3))
+			goto error;
+		break;
+		
+	case RANDOM_TYPE:
+		alive_count=calc_alive(pd,proxy_count);
+		if(alive_count<max_chain)
+			goto error_more;
+		curr_len=offset=0;
+		do {
+			if(!(p1=select_proxy(RANDOMLY,pd,proxy_count,&offset)))
+				goto error_more;
+		} while(SUCCESS!=start_chain(&ns,p1,RT) && offset<max_chain);
+		while(++curr_len<max_chain) {
+			if(!(p2=select_proxy(RANDOMLY,pd,proxy_count,&offset)))
+				goto error_more;
+			if(SUCCESS!=chain_step(ns,p1,p2))
+				goto again;
+			p1=p2;
+		}
+		proxychains_write_log(TP);
+		p3->ip=target_ip;
+		p3->port=target_port;
+		if(SUCCESS!=chain_step(ns,p1,p3))
+			goto error;
+			
 	}
 
 done:
+	proxychains_write_log("<><>-OK\n");
 	dup2(ns,sock);
 	close(ns);
 	return 0;
-error_more:
-proxychains_write_log("need more proxies\n");
 error:
 	if(ns!=-1)
 		close(ns);
-	errno=EINTR;  // for nmap ;)
+	errno = ECONNREFUSED;  // for nmap ;)
 	return -1;
+	
+error_more:
+	proxychains_write_log("\n!!!need more proxies!!!\n");
 error_strict:
-	proxychains_write_log("broken\n");
-	goto error;
+	release_all(pd,proxy_count);
+	if(ns!=-1)
+		close(ns);
+	errno = ETIMEDOUT;
+	return -1;
 }
 
+static struct hostent hostent_space;
+static in_addr_t resolved_addr;
+static char* resolved_addr_p;
+static char addr_name[1024*8];
+struct hostent* proxy_gethostbyname(const char *name)
+{
+	int pipe_fd[2];
+	char buff[256];
+	in_addr_t addr;
+	pid_t pid;
+	int status;
+	struct hostent* hp;
 
+	hostent_space.h_addr_list = &resolved_addr_p;
+	*hostent_space.h_addr_list = (char*)&resolved_addr;
+	resolved_addr = 0;
+	
+	gethostname(buff,sizeof(buff));
+	if(!strcmp(buff,name))
+		goto got_buff;
+
+	bzero(buff,sizeof(buff));
+	
+	// TODO: this works only once, so cache it  ...
+	// 	 later
+	while (hp=gethostent())
+		if (!strcmp(hp->h_name,name)) 
+			return hp; 
+	
+	if(pipe(pipe_fd))
+		goto err;
+	pid = fork();
+	switch(pid) {
+	
+		case 0: // child
+			proxychains_write_log("|DNS-request| %s \n", name);
+			dup2(pipe_fd[1],1);
+			//dup2(pipe_fd[1],2);
+		//	putenv("LD_PRELOAD=");
+			execlp("proxyresolv","proxyresolv",name,NULL);
+			perror("can't exec proxyresolv");
+			exit(2);
+
+		case -1: //error
+			close(pipe_fd[0]);
+			close(pipe_fd[1]);
+			perror("can't fork");
+			goto err;
+		
+		default:
+			close(pipe_fd[1]);
+			waitpid(pid, &status, 0);
+			read(pipe_fd[0],&buff,sizeof(buff));
+			close(pipe_fd[0]);
+got_buff:
+			addr = inet_addr(buff);
+			if (addr == -1)
+				goto err_dns;
+			memcpy(*(hostent_space.h_addr_list),
+						&addr ,sizeof(struct in_addr));
+			hostent_space.h_name = addr_name;
+			hostent_space.h_length = sizeof (in_addr_t);
+	}
+	proxychains_write_log("|DNS-response| %s is %s\n",
+			name, inet_ntoa(*(struct in_addr*)&addr));
+	return &hostent_space;
+err_dns:
+	proxychains_write_log("|DNS-response|: %s is not exist\n", name);
+err:
+	return NULL;
+}
+int proxy_getaddrinfo(const char *node, const char *service,
+		const struct addrinfo *hints,
+		struct addrinfo **res)
+{
+	struct servent *se = NULL;
+	struct hostent *hp = NULL;
+	struct sockaddr* sockaddr_space = NULL;
+	struct addrinfo*  addrinfo_space = NULL;
+	
+//	printf("proxy_getaddrinfo node %s service %s\n",node,service);
+	addrinfo_space = malloc(sizeof(struct addrinfo));
+	if(!addrinfo_space)
+		goto err1;
+	sockaddr_space = malloc(sizeof(struct sockaddr));
+	if(!sockaddr_space)
+		goto err2;
+	bzero(sockaddr_space, sizeof(*sockaddr_space));
+	bzero(addrinfo_space, sizeof(*addrinfo_space));
+	if (node &&
+	    !inet_aton(node,&((struct sockaddr_in*)sockaddr_space)->sin_addr)) {
+		hp = proxy_gethostbyname(node);
+		if (hp) 
+			memcpy(&((struct sockaddr_in*)sockaddr_space)->sin_addr,
+				*(hp->h_addr_list),
+				sizeof(in_addr_t));
+		else
+			goto err3;
+	}
+	if (service)
+		se = getservbyname(service, NULL);
+	
+	if (!se) {
+		((struct sockaddr_in*)sockaddr_space)->sin_port = 
+			htons(atoi(service?:"0"));
+	} else 
+		((struct sockaddr_in*)sockaddr_space)->sin_port = se->s_port;
+
+	*res = addrinfo_space;
+	(*res)->ai_addr = sockaddr_space;
+	if (node)
+		strcpy(addr_name, node);
+	(*res)->ai_canonname = addr_name;
+	(*res)->ai_next = NULL;
+	(*res)->ai_family = sockaddr_space->sa_family = AF_INET;
+	(*res)->ai_socktype = hints->ai_socktype;
+	(*res)->ai_flags = hints->ai_flags;
+	(*res)->ai_protocol = hints->ai_protocol;
+	(*res)->ai_addrlen = sizeof(*sockaddr_space);
+	goto out;
+err3:
+	free(sockaddr_space);
+err2:
+	free(addrinfo_space);
+err1:
+	return 1;
+out:
+	return 0;
+}
