@@ -55,21 +55,51 @@ typedef struct
 } thread_arg;
 
 int connect_proxy_chain (
-                int sock,
-                unsigned int target_ip,
-                unsigned short target_port,
-                proxy_data * pd,
-                unsigned int proxy_count,
-                chain_type ct,
-		  int max_chain );
+		int sock,
+		unsigned int target_ip,
+		unsigned short target_port,
+		proxy_data * pd,
+		unsigned int proxy_count,
+		chain_type ct,
+		int max_chain );
 
 int proxychains_write_log(char *str,...);
+struct hostent* proxy_gethostbyname(const char *name);
 
-#ifndef __linux__
-#define __libc_connect _connect
+
+typedef int (*connect_t)(int, const struct sockaddr *, socklen_t);
+connect_t true_connect;
+
+typedef struct hostent* (*gethostbyname_t)(const char *);
+gethostbyname_t true_gethostbyname;
+
+typedef int (*getaddrinfo_t)(const char *, const char *,
+		const struct addrinfo *,
+		struct addrinfo **);
+getaddrinfo_t true_getaddrinfo;
+
+typedef int (*freeaddrinfo_t)(struct addrinfo *);
+freeaddrinfo_t true_freeaddrinfo;
+
+typedef int (*getnameinfo_t) (const struct sockaddr *,
+		socklen_t, char *,
+		socklen_t, char *,
+		socklen_t, unsigned int);
+getnameinfo_t true_getnameinfo;
+
+typedef struct hostent *(*gethostbyaddr_t) (const void *, socklen_t, int);
+gethostbyaddr_t true_gethostbyaddr;
+
+int proxy_getaddrinfo(const char *node, const char *service,
+		                const struct addrinfo *hints,
+				                struct addrinfo **res);
+
+struct hostent* proxy_gethostbyname(const char *name);
+
+#if 0
+#define PDEBUG(fmt, args...) fprintf(stderr,"DEBUG:"fmt, ## args)
+#else
+#define PDEBUG(fmt, args...)
 #endif
 
-extern int __libc_connect (int sock, const struct sockaddr *addr, unsigned int len);
-
-
-#endif
+#endif 
